Module: Protocol::HTTP2::FlowControlled

Included in:
Connection, Stream
Defined in:
lib/protocol/http2/flow_controlled.rb

Overview

Provides flow control functionality for HTTP/2 connections and streams. This module implements window-based flow control as defined in RFC 7540.

Instance Method Summary collapse

Instance Method Details

#available_frame_size(maximum_frame_size = self.maximum_frame_size) ⇒ Object

This could be negative if the window has been overused due to a change in initial window size.



21
22
23
24
25
26
27
28
29
30
31
# File 'lib/protocol/http2/flow_controlled.rb', line 21

def available_frame_size(maximum_frame_size = self.maximum_frame_size)
	available_size = self.available_size
	
	# puts "available_size=#{available_size} maximum_frame_size=#{maximum_frame_size}"
	
	if available_size < maximum_frame_size
		return available_size
	else
		return maximum_frame_size
	end
end

#available_sizeObject

Get the available window size for sending data.



16
17
18
# File 'lib/protocol/http2/flow_controlled.rb', line 16

def available_size
	@remote_window.available
end

#consume_local_window(frame) ⇒ Object

Consume local window space for a received frame.



56
57
58
59
60
# File 'lib/protocol/http2/flow_controlled.rb', line 56

def consume_local_window(frame)
	# For flow-control calculations, the 9-octet frame header is not counted.
	amount = frame.length
	@local_window.consume(amount)
end

#consume_remote_window(frame) ⇒ Object

Keep track of the amount of data sent, and fail if is too much.



34
35
36
37
38
39
40
41
42
43
44
45
# File 'lib/protocol/http2/flow_controlled.rb', line 34

def consume_remote_window(frame)
	amount = frame.length
	
	# Frames with zero length with the END_STREAM flag set (that is, an empty DATA frame) MAY be sent if there is no available space in either flow-control window.
	if amount.zero? and frame.end_stream?
		# It's okay, we can send. No need to consume, it's empty anyway.
	elsif amount >= 0 and amount <= @remote_window.available
		@remote_window.consume(amount)
	else
		raise FlowControlError, "Trying to send #{frame.length} bytes, exceeded window size: #{@remote_window.available} (#{@remote_window})"
	end
end

#receive_window_update(frame) ⇒ Object

Process a received WINDOW_UPDATE frame.



82
83
84
85
86
87
88
89
90
# File 'lib/protocol/http2/flow_controlled.rb', line 82

def receive_window_update(frame)
	amount = frame.unpack
	
	if amount != 0
		@remote_window.expand(amount)
	else
		raise ProtocolError, "Invalid window size increment: #{amount}!"
	end
end

#request_window_updateObject

Request a window update if the local window is limited.



63
64
65
66
67
# File 'lib/protocol/http2/flow_controlled.rb', line 63

def request_window_update
	if @local_window.limited?
		self.send_window_update(@local_window.wanted)
	end
end

#send_window_update(window_increment) ⇒ Object

Notify the remote end that we are prepared to receive more data:



70
71
72
73
74
75
76
77
# File 'lib/protocol/http2/flow_controlled.rb', line 70

def send_window_update(window_increment)
	frame = WindowUpdateFrame.new(self.id)
	frame.pack window_increment
	
	write_frame(frame)
	
	@local_window.expand(window_increment)
end

#update_local_window(frame) ⇒ Object

Update the local window after receiving data.



49
50
51
52
# File 'lib/protocol/http2/flow_controlled.rb', line 49

def update_local_window(frame)
	consume_local_window(frame)
	request_window_update
end

#window_updated(size) ⇒ Boolean

The window has been expanded by the given amount.

Returns:

  • (Boolean)

    whether the window update was used or not.



95
96
97
# File 'lib/protocol/http2/flow_controlled.rb', line 95

def window_updated(size)
	return false
end